@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2010 Gilles Darold
+ *  Copyright (C) 2005-2012 Gilles Darold
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -40,6 +40,8 @@
 #include <errno.h>
 #include <signal.h>
 
+#define LBUFSIZ		32768
+
 /* Structure used to store information passed throught the module methods */
 typedef struct av_req_data{
      ci_simple_file_t *body;
@@ -62,13 +64,13 @@ ci_service_xdata_t *squidclamav_xdata = NULL;
 int AVREQDATA_POOL = -1;
 
 int squidclamav_init_service(ci_service_xdata_t * srv_xdata, struct ci_server_conf *server_conf);
-int squidclamav_post_init_service(ci_service_xdata_t * srv_xdata, struct ci_server_conf *server_conf);
-void squidclamav_close_service();
 int squidclamav_check_preview_handler(char *preview_data, int preview_data_len, ci_request_t *);
 int squidclamav_end_of_data_handler(ci_request_t *);
 void *squidclamav_init_request_data(ci_request_t * req);
+void squidclamav_close_service();
 void squidclamav_release_request_data(void *data);
 int squidclamav_io(char *wbuf, int *wlen, char *rbuf, int *rlen, int iseof, ci_request_t * req);
+int squidclamav_post_init_service(ci_service_xdata_t * srv_xdata, struct ci_server_conf *server_conf);
 
 /* General functions */
 void set_istag(ci_service_xdata_t * srv_xdata);
@@ -105,6 +107,7 @@ int current_pattern_size = 0;
 ci_off_t maxsize = 0;
 int logredir = 0;
 int dnslookup = 1;
+int safebrowsing = 0;
 
 /* Used by pipe to squidGuard */
 int usepipe = 0;
@@ -133,10 +136,28 @@ void cfgreload_command(char *, int, char **);
 int create_pipe(char *command);
 int dconnect (void);
 int connectINET(char *serverHost, uint16_t serverPort);
-
+char * replace(const char *s, const char *old, const char *new);
+int squidclamav_safebrowsing(ci_request_t * req, char *url, char *clientip, char *username);
 
 /* ----------------------------------------------------- */
 
+/* Sends bytes over a socket. Returns the number of bytes sent */
+int sendln(int asockd, const char *line, unsigned int len)
+{
+    int bytesent = 0;
+    while (len) {
+        int sent = send(asockd, line, len, 0);
+        if (sent <= 0) {
+            if(sent && errno == EINTR) continue;
+	    ci_debug_printf(0, "ERROR: Can't send to clamd: %s\n", strerror(errno));
+            return sent;
+        }
+        line += sent;
+        len -= sent;
+        bytesent += sent;
+    }
+    return bytesent;
+}
 
 int squidclamav_init_service(ci_service_xdata_t * srv_xdata,
                            struct ci_server_conf *server_conf)
@@ -154,14 +175,13 @@ int squidclamav_init_service(ci_service_xdata_t * srv_xdata,
     xops = CI_XCLIENTIP | CI_XSERVERIP;
     xops |= CI_XAUTHENTICATEDUSER | CI_XAUTHENTICATEDGROUPS;
     ci_service_set_xopts(srv_xdata, xops);
- 
 
     /*Initialize object pools*/
     AVREQDATA_POOL = ci_object_pool_register("av_req_data_t", sizeof(av_req_data_t));
 
     if(AVREQDATA_POOL < 0) {
 	 ci_debug_printf(0, "FATAL squidclamav_init_service: error registering object_pool av_req_data_t\n");
-	 return 0;
+	 return CI_ERROR;
     }
 
     /* Reload configuration command */
@@ -175,10 +195,10 @@ int squidclamav_init_service(ci_service_xdata_t * srv_xdata,
     memset(clamd_curr_ip, 0, sizeof(clamd_curr_ip));
 
     if (load_patterns() == 0) {
-	return 0;
+	return CI_ERROR;
     }
 
-    return 1;
+    return CI_OK;
 }
 
 void cfgreload_command(char *name, int type, char **argv)
@@ -195,6 +215,7 @@ void cfgreload_command(char *name, int type, char **argv)
     maxsize = 0;
     logredir = 0;
     dnslookup = 1;
+    safebrowsing = 0;
     clamd_curr_ip = (char *) malloc (sizeof (char) * 128);
     memset(clamd_curr_ip, 0, sizeof(clamd_curr_ip));
     if (load_patterns() == 0)
@@ -212,15 +233,18 @@ void cfgreload_command(char *name, int type, char **argv)
 int squidclamav_post_init_service(ci_service_xdata_t * srv_xdata, struct ci_server_conf *server_conf)
 {
 
-    if (squidguard == NULL) return 0;
+    if (squidguard == NULL) {
+	ci_debug_printf(1, "DEBUG squidclamav_post_init_service: squidguard not defined\n");
+	return CI_OK;
+    }
 
     ci_debug_printf(1, "DEBUG squidclamav_post_init_service: opening pipe to %s\n", squidguard);
 
     if (create_pipe(squidguard) == 1) {
-	return 0;
+	return CI_ERROR;
     }
 
-    return 1;
+    return CI_OK;
 }
 
 void squidclamav_close_service()
@@ -291,6 +315,7 @@ int squidclamav_check_preview_handler(char *preview_data, int preview_data_len,
      ci_off_t content_length;
      char *chain_ret = NULL;
      char *ret = NULL;
+     int chkipdone = 0;
 
      ci_debug_printf(1, "DEBUG squidclamav_check_preview_handler: processing preview header.\n");
 
@@ -321,18 +346,26 @@ int squidclamav_check_preview_handler(char *preview_data, int preview_data_len,
      if ((clientip = ci_headers_value(req->request_header, "X-Client-IP")) != NULL) {
 	ci_debug_printf(2, "DEBUG squidclamav_check_preview_handler: X-Client-IP: %s\n", clientip);
 	ip = inet_addr(clientip);
+	chkipdone = 0;
 	if (dnslookup == 1) {
 		if ( (clientname = gethostbyaddr((char *)&ip, sizeof(ip), AF_INET)) != NULL) {
 			if (clientname->h_name != NULL) {
-				ci_debug_printf(2, "DEBUG squidclamav_check_preview_handler: X-Client-Name: %s\n", clientname->h_name);
 				/* if a TRUSTCLIENT match => no squidguard and no virus scan */
 				if (client_pattern_compare(clientip, clientname->h_name) > 0) {
 				   ci_debug_printf(1, "DEBUG squidclamav_check_preview_handler: No squidguard and antivir check (TRUSTCLIENT match) for client: %s(%s)\n", clientname->h_name, clientip);
 				   return CI_MOD_ALLOW204;
 				}
+				chkipdone = 1;
 			}
 		  }
 	}
+	if (chkipdone == 0) {
+		/* if a TRUSTCLIENT match => no squidguard and no virus scan */
+		if (client_pattern_compare(clientip, NULL) > 0) {
+		   ci_debug_printf(1, "DEBUG squidclamav_check_preview_handler: No squidguard and antivir check (TRUSTCLIENT match) for client: %s\n", clientip);
+		   return CI_MOD_ALLOW204;
+		}
+	}
      } else {
 	/* set null client to - */
 	clientip = (char *)malloc(sizeof(char)*2);
@@ -356,10 +389,14 @@ int squidclamav_check_preview_handler(char *preview_data, int preview_data_len,
      
      /* Check URL header against squidGuard */
      if (usepipe == 1) {
+	char *rbuff = NULL;
 	ci_debug_printf(2, "DEBUG squidclamav_check_preview_handler: Sending request to chained program: %s\n", squidguard);
 	ci_debug_printf(2, "DEBUG squidclamav_check_preview_handler: Request: %s %s %s %s\n", httpinf.url,clientip,username,httpinf.method);
-	fprintf(sgfpw,"%s %s %s %s\n",httpinf.url,clientip,username,httpinf.method);
+	/* escaping escaped character to prevent unescaping by squidguard */
+	rbuff = replace(httpinf.url, "%", "%25");
+	fprintf(sgfpw,"%s %s %s %s\n",rbuff,clientip,username,httpinf.method);
 	fflush(sgfpw);
+	xfree(rbuff);
 	/* the chained redirector must return empty line if ok or the redirection url */
 	chain_ret = (char *)malloc(sizeof(char)*MAX_URL_SIZE);
 	if (chain_ret != NULL) {
@@ -392,14 +429,19 @@ int squidclamav_check_preview_handler(char *preview_data, int preview_data_len,
 	   return CI_MOD_ALLOW204;
      }
 
+     if (safebrowsing == 1) {
+	if (squidclamav_safebrowsing(req, httpinf.url, clientip, username) != 0) {
+		ci_debug_printf(1, "DEBUG squidclamav_check_preview_handler: Malware found stopping here.\n");
+	        return CI_MOD_CONTINUE;
+	 }
+     }
      /* Get the content length header */
-     if ((content_length = ci_http_content_length(req)) > 0) {
-	ci_debug_printf(2, "DEBUG squidclamav_check_preview_handler: Content-Length: %d\n", (int)content_length);
-        /* check the Content-Length against maxsize, if upper no scan */
-        if ((maxsize > 0) && (content_length >= maxsize)) {
-	   ci_debug_printf(1, "DEBUG squidclamav_check_preview_handler: No antivir check, Content-Length is upper than maxsize (%d>%d)\n", (int)content_length, (int)maxsize);
-	   return CI_MOD_ALLOW204;
-        }
+     content_length = ci_http_content_length(req);
+     ci_debug_printf(2, "DEBUG squidclamav_check_preview_handler: Content-Length: %d\n", (int)content_length);
+
+     if ((content_length > 0) && (maxsize > 0) && (content_length >= maxsize)) {
+	ci_debug_printf(2, "DEBUG squidclamav_check_preview_handler: No antivir check, content-length upper than maxsize (%d > %d)\n", (int)content_length, (int)maxsize);
+	return CI_MOD_ALLOW204;
      }
 
      /* Get the content type header */
@@ -439,8 +481,8 @@ int squidclamav_check_preview_handler(char *preview_data, int preview_data_len,
 	data->clientip = NULL;
      }
 
-     data->body = ci_simple_file_new(maxsize);
-     if (SEND_PERCENT_BYTES >= 0 && START_SEND_AFTER == 0) {
+     data->body = ci_simple_file_new(0);
+     if ((SEND_PERCENT_BYTES >= 0) && (START_SEND_AFTER == 0)) {
 	ci_req_unlock_data(req);
 	ci_simple_file_lock_all(data->body);
      }
@@ -470,11 +512,12 @@ int squidclamav_read_from_net(char *buf, int len, int iseof, ci_request_t * req)
 	return ci_simple_file_write(data->body, buf, len, iseof);
     }
 
-    if ((maxsize > 0) && (ci_simple_file_size(data->body) >= maxsize)) {
+    if ((maxsize > 0) && (data->body->bytes_in >= maxsize)) {
 	data->no_more_scan = 1;
 	ci_req_unlock_data(req);
 	ci_simple_file_unlock_all(data->body);
-    } else if (SEND_PERCENT_BYTES && START_SEND_AFTER < ci_simple_file_size(data->body)) {
+	ci_debug_printf(1, "DEBUG squidclamav_read_from_net: No more antivir check, downloaded stream is upper than maxsize (%d>%d)\n", (int)data->body->bytes_in, (int)maxsize);
+    } else if (SEND_PERCENT_BYTES && (START_SEND_AFTER < data->body->bytes_in)) {
 	ci_req_unlock_data(req);
 	allow_transfer = (SEND_PERCENT_BYTES * (data->body->endpos + len)) / 100;
 	ci_simple_file_unlock(data->body, allow_transfer);
@@ -542,14 +585,7 @@ int squidclamav_end_of_data_handler(ci_request_t * req)
 
      ssize_t ret;
      int nbread = 0;
-     int loopw = 60;
-     uint16_t port;
-     struct sockaddr_in server;
-     struct sockaddr_in peer;
-     size_t peer_size;
-     char *pt = NULL;
      int sockd;
-     int wsockd; 
      unsigned long total_read;
 
      ci_debug_printf(2, "DEBUG squidclamav_end_of_data_handler: ending request data handler.\n");
@@ -573,71 +609,17 @@ int squidclamav_end_of_data_handler(ci_request_t * req)
      /* SCAN DATA HERE */
      if ((sockd = dconnect ()) < 0) {
 	ci_debug_printf(0, "ERROR squidclamav_end_of_data_handler: Can't connect to Clamd daemon.\n");
-	return CI_MOD_ALLOW204;
+	goto done_allow204;
      }
-     ci_debug_printf(1, "DEBUG squidclamav_end_of_data_handler: Sending STREAM command to clamd.\n");
+     ci_debug_printf(1, "DEBUG squidclamav_end_of_data_handler: Sending zINSTREAM command to clamd.\n");
 
-     if (write(sockd, "STREAM", 6) <= 0) {
+     if (write(sockd, "zINSTREAM", 10) <= 0) {
 	ci_debug_printf(0, "ERROR squidclamav_end_of_data_handler: Can't write to Clamd socket.\n");
 	close(sockd);
-	return CI_MOD_ALLOW204;
+	goto done_allow204;
      }
 
-     while (loopw > 0) {
-	memset (cbuff, 0, sizeof(cbuff));
-	ret = read (sockd, cbuff, MAX_URL_SIZE);
-	if ((ret > -1) && (pt = strstr (cbuff, "PORT"))) {
-	   pt += 5;
-	   sscanf(pt, "%d", (int *) &port);
-	   break;
-	}
-	loopw--;
-     }
-     if (loopw == 0) {
-	ci_debug_printf(0, "ERROR squidclamav_end_of_data_handler: Clamd daemon not ready for stream scanning.\n");
-	close(sockd);
-	return CI_MOD_ALLOW204;
-     }
-
-     ci_debug_printf(1, "DEBUG squidclamav_end_of_data_handler: Received port %d from clamd.\n", port);
-
-     /* connect to clamd given port */
-     if ((wsockd = socket (AF_INET, SOCK_STREAM, 0)) < 0) {
-	ci_debug_printf(0, "ERROR squidclamav_end_of_data_handler: Can't create the Clamd socket.\n");
-	close(sockd);
-	return CI_MOD_ALLOW204;
-     }
-
-     server.sin_family = AF_INET;
-     server.sin_port = htons (port);
-     peer_size = sizeof (peer);
-
-     if (getpeername(sockd, (struct sockaddr *) &peer, (socklen_t *) &peer_size) < 0) {
-	ci_debug_printf(0, "ERROR squidclamav_end_of_data_handler: Can't get socket peer name.\n");
-	close(sockd);
-	return CI_MOD_ALLOW204;
-     }
-     switch (peer.sin_family) {
-	case AF_UNIX:
-	server.sin_addr.s_addr = inet_addr ("127.0.0.1");
-	break;
-	case AF_INET:
-	server.sin_addr.s_addr = peer.sin_addr.s_addr;
-	break;
-	default:
-	ci_debug_printf(0, "ERROR squidclamav_end_of_data_handler: Unexpected socket type: %d.\n", peer.sin_family);
-	close(sockd);
-	return CI_MOD_ALLOW204;
-     }
-
-     ci_debug_printf(1, "DEBUG squidclamav_end_of_data_handler: Trying to connect to clamd [port: %d].\n", port);
-
-     if (connect (wsockd, (struct sockaddr *) &server, sizeof (struct sockaddr_in)) < 0) {
-	close(wsockd);
-	ci_debug_printf(0, "ERROR squidclamav_end_of_data_handler: Can't connect to clamd [port: %d].\n", port);
-	return CI_MOD_ALLOW204;
-     }
-     ci_debug_printf(1, "DEBUG squidclamav_end_of_data_handler: Ok connected to clamd on port: %d.\n", port);
+     ci_debug_printf(1, "DEBUG squidclamav_end_of_data_handler: Ok connected to clamd.\n");
 
 /*-----------------------------------------------------*/
 
@@ -645,9 +627,12 @@ int squidclamav_end_of_data_handler(ci_request_t * req)
      lseek(body->fd, 0, SEEK_SET);
      memset(cbuff, 0, sizeof(cbuff));
      total_read = 0;
-     while (data->virus == 0 && (nbread = read(body->fd, cbuff, MAX_URL_SIZE)) > 0) {
+     while (data->virus == 0 && (nbread = read(body->fd, cbuff, BUFSIZ)) > 0) {
+	    uint32_t buf[LBUFSIZ/sizeof(uint32_t)];
+	    buf[0] = htonl(nbread);
+	    memcpy(&buf[1],(const char*) cbuff, nbread);
 	    total_read += nbread;
-	    ret = write(wsockd, cbuff, nbread);
+	    ret = sendln (sockd,(const char *) buf, nbread+sizeof(uint32_t));
 	    if ( (ret <= 0) && (total_read > 0) ) {
 		ci_debug_printf(3, "ERROR squidclamav_end_of_data_handler: Can't write to clamd socket (maybe we reach clamd StreamMaxLength, total read: %ld).\n", total_read);
 		break;
@@ -657,38 +642,41 @@ int squidclamav_end_of_data_handler(ci_request_t * req)
 	    } else {
 		ci_debug_printf(3, "DEBUG squidclamav_end_of_data_handler: Write %d bytes on %d to socket\n", (int)ret, nbread);
 	    }
-
 	    memset(cbuff, 0, sizeof(cbuff));
-
      }
 
-     /* close socket to clamd */
-     if (wsockd > -1) {
-        ci_debug_printf(1, "DEBUG squidclamav_end_of_data_handler: End Clamd connection, attempting to read result.\n");
-	close(wsockd);
+     uint32_t buf[LBUFSIZ/sizeof(uint32_t)];
+     *buf = 0;
+     ret = sendln (sockd,(const char *) buf, 4);
+     if (ret <= 0)
+     {
+       ci_debug_printf(0, "ERROR squidclamav_end_of_data_handler: Can't write zINSTREAM ending chars to clamd socket.\n");
+     } else {
+
+	     /* Reading clamd result */
+	     memset (clbuf, 0, sizeof(clbuf));
+	     while ((nbread = read(sockd, clbuf, SMALL_BUFF)) > 0) {
+		ci_debug_printf(1, "DEBUG squidclamav_end_of_data_handler: received from Clamd: %s\n", clbuf);
+		if (strstr (clbuf, "FOUND")) {
+		   data->virus = 1;
+		   if (!ci_req_sent_data(req)) {
+			chomp(clbuf);
+			char *urlredir = (char *) malloc( sizeof(char)*MAX_URL_SIZE );
+			snprintf(urlredir, MAX_URL_SIZE, "%s?url=%s&source=%s&user=%s&virus=%s", redirect_url, data->url, data->clientip, data->user, clbuf);
+			if (logredir == 0)
+			   ci_debug_printf(1, "DEBUG squidclamav_end_of_data_handler: Virus redirection: %s.\n", urlredir);
+			if (logredir)
+			    ci_debug_printf(0, "INFO squidclamav_end_of_data_handler: Virus redirection: %s.\n", urlredir);
+			generate_redirect_page(urlredir, req, data);
+			xfree(urlredir);
+		   }
+		   ci_debug_printf(1, "DEBUG squidclamav_end_of_data_handler: Virus found, ending download.\n");
+		   break;
+		}
+		memset(clbuf, 0, sizeof(clbuf));
+	     }
      }
 
-     memset (clbuf, 0, sizeof(clbuf));
-     while ((nbread = read(sockd, clbuf, SMALL_BUFF)) > 0) {
-	ci_debug_printf(1, "DEBUG squidclamav_end_of_data_handler: received from Clamd: %s", clbuf);
-	if (strstr (clbuf, "FOUND\n")) {
-	   data->virus = 1;
-	   if (!ci_req_sent_data(req)) {
-		chomp(clbuf);
-		char *urlredir = (char *) malloc( sizeof(char)*MAX_URL_SIZE );
-		snprintf(urlredir, MAX_URL_SIZE, "%s?url=%s&source=%s&user=%s&virus=%s", redirect_url, data->url, data->clientip, data->user, clbuf);
-		if (logredir == 0)
-		   ci_debug_printf(1, "DEBUG squidclamav_end_of_data_handler: Virus redirection: %s.\n", urlredir);
-		if (logredir)
-		    ci_debug_printf(0, "INFO squidclamav_end_of_data_handler: Virus redirection: %s.\n", urlredir);
-		generate_redirect_page(urlredir, req, data);
-		xfree(urlredir);
-	   }
-	   ci_debug_printf(1, "DEBUG squidclamav_end_of_data_handler: Virus found, ending download.\n");
-	   break;
-	}
-	memset(clbuf, 0, sizeof(clbuf));
-     }
      /* close second socket to clamd */
      if (sockd > -1) {
         ci_debug_printf(1, "DEBUG squidclamav_end_of_data_handler: Closing Clamd connection.\n");
@@ -700,7 +688,8 @@ int squidclamav_end_of_data_handler(ci_request_t * req)
           return CI_MOD_DONE;
      }
 
-     if (!ci_req_sent_data(req)) {
+done_allow204:
+     if (!ci_req_sent_data(req) && ci_req_allow204(req)) {
 	ci_debug_printf(2, "DEBUG squidclamav_end_of_data_handler: Responding with allow 204\n");
 	return CI_MOD_ALLOW204;
      }
@@ -996,7 +985,7 @@ client_pattern_compare(char *ip, char *name)
 				return 1;
 			/* Look at client name pattern matching */
 			/* return 2 if string matches fqdn TRUSTCLIENT pattern */
-			} else if (regexec(&patterns[i].regexv, name, 0, 0, 0) == 0) {
+			} else if ((name != NULL) && (regexec(&patterns[i].regexv, name, 0, 0, 0) == 0)) {
 				if (debug != 0)
 					ci_debug_printf(2, "DEBUG client_pattern_compare: trustclient (%s) matched: %s\n", patterns[i].pattern, name);
 				return 2;
@@ -1192,6 +1181,14 @@ add_pattern(char *s)
 		return 1;
 	}
 
+	if(strcmp(type, "safebrowsing") == 0) {
+		if (safebrowsing == 0)
+		   safebrowsing = atoi(first);
+		xfree(type);
+		xfree(first);
+		return 1;
+	}
+
 	if(strcmp(type, "timeout") == 0) {
 		timeout = atoi(first);
 		if (timeout > 10)
@@ -1619,3 +1616,123 @@ connectINET(char *serverHost, uint16_t serverPort)
         return asockd;
 }
 
+
+/**
+ * Searches all occurrences of old into s
+ * and replaces with new
+ */
+char *
+replace(const char *s, const char *old, const char *new)
+{
+	char *ret;
+	int i, count = 0;
+	size_t newlen = strlen(new);
+	size_t oldlen = strlen(old);
+
+	for (i = 0; s[i] != '\0'; i++) {
+		if (strstr(&s[i], old) == &s[i]) {
+			count++;
+			i += oldlen - 1;
+		}
+	}
+	ret = malloc(i + 1 + count * (newlen - oldlen));
+	if (ret != NULL) {
+		i = 0;
+		while (*s) {
+			if (strstr(s, old) == s) {
+				strcpy(&ret[i], new);
+				i += newlen;
+				s += oldlen;
+			} else {
+				ret[i++] = *s++;
+			}
+		}
+		ret[i] = '\0';
+	}
+
+	return ret;
+}
+
+int
+squidclamav_safebrowsing(ci_request_t * req, char *url, char *clientip, char *username)
+{
+     av_req_data_t *data = ci_service_data(req);
+     char cbuff[MAX_URL_SIZE+60];
+     char clbuf[SMALL_BUFF];
+
+     ssize_t ret;
+     int nbread = 0;
+     int sockd;
+
+     ci_debug_printf(2, "DEBUG squidclamav_safebrowsing: looking for Clamav SafeBrowsing check.\n");
+
+     /* SCAN DATA HERE */
+     if ((sockd = dconnect ()) < 0) {
+	ci_debug_printf(0, "ERROR squidclamav_safebrowsing: Can't connect to Clamd daemon.\n");
+	return 0;
+     }
+     ci_debug_printf(1, "DEBUG squidclamav_safebrowsing: Sending zINSTREAM command to clamd.\n");
+
+     if (write(sockd, "zINSTREAM", 10) <= 0) {
+	ci_debug_printf(0, "ERROR squidclamav_safebrowsing: Can't write to Clamd socket.\n");
+	close(sockd);
+	return 0;
+     }
+
+     ci_debug_printf(1, "DEBUG squidclamav_safebrowsing: Ok connected to clamd socket.\n");
+
+     ci_debug_printf(1, "DEBUG: squidclamav_safebrowsing: Scanning url for Malware now\n");
+     uint32_t buf[BUFSIZ/sizeof(uint32_t)];
+     strcpy(cbuff, "From test\n\n<a href=");
+     strcat(cbuff, url);
+     strcat(cbuff, ">squidclamav-safebrowsing-test</a>\n");
+     size_t sfsize = 0;
+     sfsize = strlen(cbuff);
+     buf[0] = htonl(sfsize);
+     memcpy(&buf[1],(const char*) cbuff, sfsize);
+     ci_debug_printf(3, "DEBUG: squidclamav_safebrowsing: sending %s\n", cbuff);
+     ret = sendln (sockd,(const char *) buf, sfsize+sizeof(uint32_t));
+     if ( ret <= 0 ) {
+	ci_debug_printf(0, "ERROR squidclamav_safebrowsing: Can't write to clamd socket.\n");
+     } else {
+	     ci_debug_printf(3, "DEBUG squidclamav_safebrowsing: Write to socket\n");
+	     memset(cbuff, 0, sizeof(cbuff));
+	     *buf = 0;
+	     ret = sendln (sockd,(const char *) buf, 4);
+	     if (ret <= 0)
+	     {
+		ci_debug_printf(0, "ERROR squidclamav_safebrowsing: Can't write INSTREAM ending chars to clamd socket.\n");
+	     } else {
+		     memset (clbuf, 0, sizeof(clbuf));
+		     while ((nbread = read(sockd, clbuf, SMALL_BUFF)) > 0) {
+			ci_debug_printf(1, "DEBUG squidclamav_safebrowsing: received from Clamd: %s\n", clbuf);
+			if (strstr (clbuf, "FOUND")) {
+				char *urlredir = (char *) malloc( sizeof(char)*MAX_URL_SIZE );
+				chomp(clbuf);
+				snprintf(urlredir, MAX_URL_SIZE, "%s?url=%s&source=%s&user=%s&malware=%s", redirect_url, url, clientip, username, clbuf);
+				if (logredir == 0)
+				   ci_debug_printf(1, "DEBUG squidclamav_safebrowsing: Malware redirection: %s.\n", urlredir);
+				if (logredir)
+				    ci_debug_printf(0, "INFO squidclamav_safebrowsing: Malware redirection: %s.\n", urlredir);
+				/* Create the redirection url to squid */
+				data->blocked = 1;
+				generate_redirect_page(urlredir, req, data);
+				xfree(urlredir);
+				return 1;
+			}
+			memset(clbuf, 0, sizeof(clbuf));
+		     }
+		}
+     }
+     /* close socket to clamd */
+     if (sockd > -1) {
+        ci_debug_printf(1, "DEBUG squidclamav_safebrowsing: Closing Clamd connection.\n");
+	close(sockd);
+     }
+
+     ci_debug_printf(3, "DEBUG squidclamav_safebrowsing: No malware found.\n");
+
+     return 0;
+}
+
+
